from __future__ import annotations

import base64
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Literal, Protocol, TypeVar

from pydantic_core import PydanticCustomError, core_schema
from pydantic_core.core_schema import CoreSchema, WhenUsed

if TYPE_CHECKING:
    from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
    from pydantic.json_schema import JsonSchemaValue
    from pydantic_core.core_schema import SerializerFunctionWrapHandler


#: Generic type used for our encoders.
T = TypeVar('T', bound=bytes | int)


class EncoderProtocol(Protocol, Generic[T]):
    """Protocol for encoding from and decoding data to another type."""

    @classmethod
    def decode(cls, data: str) -> T:
        """
        Decode the data using the encoder.

        Args:
            data: A string that can be decoded to type ``T``.

        Returns:
            The newly decoded type.

        """

    @classmethod
    def encode(cls, value: T) -> str:
        """
        Encode the provided value using the encoder.

        Args:
            value: A generic value of type ``T``.

        Returns:
            The exact value encoded to a string.

        """

    @classmethod
    def get_json_format(cls) -> str:
        """
        Get the JSON format for the encoded data.

        Returns:
            A short descriptive name for the format.

        """


class Base32Encoder(EncoderProtocol[bytes]):
    """Encoder for base32 bytes."""

    #: Whether we are case insensitive when decoding.
    casefold: ClassVar[bool] = True
    #: Whether to remove the padding characters when serializing.
    trim_padding: ClassVar[bool] = True

    @classmethod
    def decode(cls, data: str) -> bytes:
        """
        Decode the provided base32 bytes to original bytes data.

        Args:
            data: A base32-encoded string to decode.

        Raises:
            PydanticCustomError: On decoding error.

        Returns:
            The corresponding decoded bytes.

        """
        try:
            if cls.trim_padding:
                data = data.rstrip('=')
                padlen = -len(data) % 8
                data = data + padlen * '='
            return base64.b32decode(data, cls.casefold)
        except ValueError as e:
            raise PydanticCustomError(
                'base32_decode',
                "Base32 decoding error: '{error}'",
                {'error': str(e)},
            ) from e

    @classmethod
    def encode(cls, value: bytes) -> str:
        """
        Encode a value to a base32 encoded string.

        Args:
            value: A byte value to encode to base32.

        Returns:
            The corresponding encoded string value.

        """
        encoded = base64.b32encode(value)
        if cls.trim_padding:
            encoded = encoded.rstrip(b'=')
        return encoded.decode()

    @classmethod
    def get_json_format(cls) -> Literal['base32']:
        """Get the JSON format for the encoded data."""
        return 'base32'


class Base64Encoder(EncoderProtocol[bytes]):
    """Encoder for base64 bytes."""

    #: Whether to remove the padding characters when serializing.
    trim_padding: ClassVar[bool] = True

    @classmethod
    def decode(cls, data: str) -> bytes:
        """
        Decode the provided base64 bytes to original bytes data.

        Args:
            data: A base64-encoded string to decode.

        Raises:
            PydanticCustomError: On decoding error.

        Returns:
            The corresponding decoded bytes.

        """
        try:
            encoded = data.encode()
            if cls.trim_padding:
                encoded = encoded.rstrip(b'=')
                padlen = -len(encoded) % 4
                encoded = encoded + padlen * b'='
            return base64.standard_b64decode(encoded)
        except ValueError as e:
            raise PydanticCustomError(
                'base64_decode',
                "Base64 decoding error: '{error}'",
                {'error': str(e)},
            ) from e

    @classmethod
    def encode(cls, value: bytes) -> str:
        """
        Encode a value to a base64 encoded string.

        Args:
            value: A byte value to encode to base64.

        Returns:
            The corresponding encoded string value.

        """
        encoded = base64.standard_b64encode(value)
        if cls.trim_padding:
            encoded = encoded.rstrip(b'=')
        return encoded.decode()

    @classmethod
    def get_json_format(cls) -> Literal['base64']:
        """Get the JSON format for the encoded data."""
        return 'base64'


class Base16Encoder(EncoderProtocol[bytes]):
    """Specific encoder for hex encoded strings."""

    @classmethod
    def decode(cls, data: str) -> bytes:
        """
        Decode the provided hex string to original bytes data.

        Args:
            data: A hex-encoded string to decode.

        Raises:
            PydanticCustomError: On decoding error.

        Returns:
            The corresponding decoded bytes.

        """
        try:
            return bytes.fromhex(data.zfill((len(data) + 1) & ~1))
        except ValueError as e:
            raise PydanticCustomError(
                'base16_decode',
                "Base16 decoding error: '{error}'",
                {'error': str(e)},
            ) from e

    @classmethod
    def encode(cls, value: bytes) -> str:
        """
        Encode a value to a hex encoded string.

        Args:
            value: A byte value to encode to a hexadecimal string.

        Returns:
            The corresponding encoded string value.

        """
        return value.hex()

    @classmethod
    def get_json_format(cls) -> Literal['base16']:
        """Get the JSON format for the encoded data."""
        return 'base16'


@dataclass(frozen=True, slots=True)
class EncodedBase(Generic[T]):
    """Generic encoded value to/from a string using the :class:`EncoderProtocol`."""

    #: Main core validation schema.
    CORE_SCHEMA: ClassVar[CoreSchema]
    #: Core python type associated with the schema.
    CORE_TYPE: ClassVar[type[Any]]

    #: The encoder protocol to use.
    encoder: type[EncoderProtocol[T]]
    #: When to use the encoder.
    when_used: WhenUsed = 'always'

    def __get_pydantic_core_schema__(
        self,
        source: type[Any],
        handler: GetCoreSchemaHandler,
    ) -> CoreSchema:
        """Tell the core schema and how to validate the whole thing."""
        return core_schema.no_info_before_validator_function(
            function=self._pydantic_validator,
            schema=handler(source),
            # We are bytes and will be returned as a string.
            serialization=core_schema.wrap_serializer_function_ser_schema(
                function=self._pydantic_serializer,
                schema=handler(source),
                when_used=self.when_used,
            ),
        )

    def __get_pydantic_json_schema__(
        self,
        schema: CoreSchema,
        handler: GetJsonSchemaHandler,
    ) -> JsonSchemaValue:
        """
        Update JSON schema to also tell about this field.

        See Also:
            https://json-schema.org/draft/2020-12/json-schema-validation#name-contentencoding

        """
        content_encoding = self.encoder.get_json_format()
        field_schema = handler(self.CORE_SCHEMA)
        field_schema.update(contentEncoding=content_encoding, type='string')
        return field_schema

    def _pydantic_serializer(self, item: Any, serialize: SerializerFunctionWrapHandler) -> str:
        """Serialize the provided item to a string."""
        # This is needed because bytes are serialized to strings by pydantic :(.
        if isinstance(item, self.CORE_TYPE):
            return self.to_string(item)
        return self.to_string(serialize(item))

    def _pydantic_validator(self, value: Any) -> Any:
        """Validate any kind of input data."""
        if isinstance(value, str):
            return self.from_string(value)
        return value

    def from_string(self, string: str) -> T:
        """Decode a string to the underlying type."""
        return self.encoder.decode(string)

    def to_string(self, value: T) -> str:
        """Encode the value using the specified encoder."""
        return self.encoder.encode(value)


@dataclass(frozen=True, slots=True)
class EncodedBytes(EncodedBase[bytes]):
    """Bytes that can be encoded and decoded from a string using an external encoder."""

    #: Our core schema is for :class:`bytes`.
    CORE_SCHEMA: ClassVar[CoreSchema] = core_schema.bytes_schema(strict=True)

    #: Core python type associated with the schema.
    CORE_TYPE: ClassVar[type[Any]] = bytes

    def __hash__(self) -> int:
        """
        Provide the hash from the encoder.

        Returns:
            An unique hash for our byte encoder.

        """
        return hash(self.encoder)



from __future__ import annotations

import asyncio
import dataclasses

import pytest

from aiostem import ControllerStatus, Monitor
from aiostem.event import Event, EventNetworkLiveness, EventStatusClient, event_from_message
from aiostem.utils import Message

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class FakeMonitor(Monitor):
    """Provide a direct access to internal methods to make mypy happy."""

    event_until_healthy = asyncio.Event()
    event_for_error = asyncio.Event()

    async def on_ctrl_client_status(self, event: EventStatusClient) -> None:
        await self._on_ctrl_client_status(event)

    async def on_ctrl_liveness_status(self, event: EventNetworkLiveness) -> None:
        await self._on_ctrl_liveness_status(event)

    async def wait_until_healthy(self) -> ControllerStatus:
        self.event_until_healthy.set()
        return await super().wait_until_healthy()

    async def wait_for_error(self) -> ControllerStatus:
        self.event_for_error.set()
        return await super().wait_for_error()


def build_event(line: str) -> Event:
    return event_from_message(Message(status=650, header=line))


class TestMonitor:
    @pytest.mark.timeout(5)
    async def test_monitor_not_entered(self, controller):
        monitor = Monitor(controller)
        assert monitor.is_healthy is False
        assert monitor.is_entered is False

        status = await monitor.wait_for_error()
        assert isinstance(status, ControllerStatus), type(status)
        assert status.is_healthy is False
        assert status == monitor.status

        await monitor.__aexit__(None, None, None)
        assert monitor.is_entered is False

    @pytest.mark.timeout(5)
    async def test_monitor_entered(self, controller):
        async with Monitor(controller) as monitor:
            with pytest.raises(RuntimeError, match='is already running'):
                await monitor.__aenter__()

    @pytest.mark.timeout(5)
    async def test_monitor_keepalive_signal(self, controller):
        controller.traces.add('signal')

        async with Monitor(controller, keepalive=True):
            await controller.event_signal_active.wait()
            assert 'ACTIVE' in controller.trace_signals

    @pytest.mark.timeout(5)
    async def test_monitor_events(self, controller):
        monitor = FakeMonitor(controller)

        # Check that we can properly handle PROGRESS messages sent by the controller.
        assert monitor.status.bootstrap == 0
        event = build_event(
            'STATUS_CLIENT NOTICE BOOTSTRAP TAG=done SUMMARY="Done" PROGRESS=100',
        )
        assert isinstance(event, EventStatusClient)
        await monitor.on_ctrl_client_status(event)
        assert monitor.status.bootstrap == 100

        # Check that we can properly handle CIRCUIT_ESTABLISHED events.
        assert monitor.status.has_circuits is False
        event = build_event('STATUS_CLIENT NOTICE CIRCUIT_ESTABLISHED')
        await monitor.on_ctrl_client_status(event)
        assert monitor.status.has_circuits is True

        # Check that we can properly handle ENOUGH_DIR_INFO events.
        assert monitor.status.has_dir_info is False
        event = build_event('STATUS_CLIENT NOTICE ENOUGH_DIR_INFO')
        await monitor.on_ctrl_client_status(event)
        assert monitor.status.has_dir_info is True

        # We now have enough to tell that we are in a healthy state.
        assert monitor.is_healthy is True
        async with monitor.condition:
            await monitor.condition.wait_for(lambda: monitor.is_healthy)
        assert monitor.status.is_healthy is True

        assert monitor.status.net_liveness is False
        event = build_event('NETWORK_LIVENESS UP')
        assert isinstance(event, EventNetworkLiveness)
        await monitor.on_ctrl_liveness_status(event)
        assert monitor.status.net_liveness is True
        assert monitor.is_healthy is True

        event = build_event('STATUS_CLIENT NOTICE CIRCUIT_NOT_ESTABLISHED REASON=CLOCK_JUMPED')
        await monitor.on_ctrl_client_status(event)
        assert monitor.status.has_circuits is False
        assert monitor.is_healthy is True

        event = build_event('STATUS_CLIENT NOTICE NOT_ENOUGH_DIR_INFO')
        await monitor.on_ctrl_client_status(event)
        assert monitor.status.has_dir_info is False
        assert monitor.is_healthy is False

        old_status = dataclasses.replace(monitor.status)
        event = build_event('STATUS_CLIENT NOTICE CONSENSUS_ARRIVED')
        await monitor.on_ctrl_client_status(event)
        assert monitor.status == old_status

    @pytest.mark.timeout(5)
    async def test_monitor_wait_methods(self, controller):
        monitor = FakeMonitor(controller)
        monitor.status.bootstrap = 100
        monitor.status.has_dir_info = True

        assert monitor.is_healthy is False
        task = asyncio.create_task(monitor.wait_until_healthy())
        await monitor.event_until_healthy.wait()
        await asyncio.sleep(0.05)

        event = build_event('NETWORK_LIVENESS UP')
        assert isinstance(event, EventNetworkLiveness)
        await monitor.on_ctrl_liveness_status(event)
        status = await task
        assert status.is_healthy is True, status

        task = asyncio.create_task(monitor.wait_for_error())
        await monitor.event_for_error.wait()
        await asyncio.sleep(0.05)

        event = build_event('NETWORK_LIVENESS DOWN')
        assert isinstance(event, EventNetworkLiveness)
        await monitor.on_ctrl_liveness_status(event)
        status = await task
        assert status.is_healthy is False, status



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Any, ClassVar

import pytest
from pydantic import BaseModel, TypeAdapter, ValidationError

from aiostem.types import Base16Bytes, Base32Bytes, Base64Bytes
from aiostem.utils import Base32Encoder, Base64Encoder, EncodedBytes

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


class BaseEncoderTest:
    DECODED_VALUE = b'These are bytes!'
    ENCODED_VALUE = ''
    TEST_CLASS = NotImplemented
    SCHEMA_FORMAT = 'format'
    VALUES: ClassVar[Mapping[str, Sequence[Any]]] = {
        'good': [],
        'fail': [],
    }

    def stub_fail_values(self, value):
        message = f'{self.SCHEMA_FORMAT.capitalize()} decoding error:'
        with pytest.raises(ValidationError, match=message):
            self.TEST_MODEL(v=value)

    def stub_good_values(self, value):
        model = self.TEST_MODEL(v=value)
        assert model.v == self.DECODED_VALUE

    def stub_good_encoding(self, value):
        model = self.TEST_MODEL(v=value)
        assert model.model_dump_json() == '{"v":"' + self.ENCODED_VALUE + '"}'

    def test_passthough(self):
        model = self.TEST_MODEL_OR_INT(v=123)
        assert model.model_dump_json() == '{"v":123}'

    def test_json_schema(self):
        schema = self.TEST_MODEL.model_json_schema()
        if self.SCHEMA_FORMAT is not None:
            assert schema['properties']['v'] == {
                'contentEncoding': self.SCHEMA_FORMAT,
                'format': 'binary',
                'title': 'V',
                'type': 'string',
            }
        else:
            assert schema['properties']['v'] == {'title': 'V', 'type': 'string'}


# Dirty decorator to make our tests dynamic.
# This looks for all 'stub_' methods in our direct parent and wraps this function
# around pytest.mark.parametrize to inject our test values.
def inject_test_values(cls):
    class TestModel(BaseModel):
        v: cls.TEST_CLASS

    class TestModelOrInt(BaseModel):
        v: cls.TEST_CLASS | int

    for name, method in BaseEncoderTest.__dict__.items():
        if name.startswith('stub_'):
            action = name.split('_')[1]
            values = [(method, value) for value in cls.VALUES.get(action, [])]

            @pytest.mark.parametrize(('method', 'value'), values)
            def wrapper(self, method, value):
                return method(self, value)

            setattr(cls, 'test_' + name[5:], wrapper)

    cls.TEST_MODEL = TestModel
    cls.TEST_MODEL_OR_INT = TestModelOrInt
    cls.ADAPTER = TypeAdapter(cls.TEST_CLASS)

    return cls


@inject_test_values
class TestBase16Bytes(BaseEncoderTest):
    TEST_CLASS = Base16Bytes
    ENCODED_VALUE = '54686573652061726520627974657321'
    SCHEMA_FORMAT = 'base16'
    VALUES: ClassVar[Mapping[str, Sequence[Any]]] = {
        'good': [
            b'These are bytes!',
            Base16Bytes(b'These are bytes!'),
            '54686573652061726520627974657321',
        ],
        'fail': ['54T6'],
    }


class Base32PaddedEncoder(Base32Encoder):
    trim_padding: ClassVar[bool] = False


Base32BytesPadded = Annotated[bytes, EncodedBytes(encoder=Base32PaddedEncoder)]


@inject_test_values
class TestBase32(BaseEncoderTest):
    TEST_CLASS = Base32Bytes
    ENCODED_VALUE = 'KRUGK43FEBQXEZJAMJ4XIZLTEE'
    SCHEMA_FORMAT = 'base32'
    VALUES: ClassVar[Mapping[str, Sequence[Any]]] = {
        'fail': [
            'KRUGK43FEBQXEZJAMJ4XIZLTE9',  # Invalid character
        ],
        'good': [
            b'These are bytes!',
            Base32Bytes(b'These are bytes!'),
            'KRUGK43FEBQXEZJAMJ4XIZLTEE',
            'krugk43febqxezjamj4xizltee',
        ],
    }


@inject_test_values
class TestBase32Padded(BaseEncoderTest):
    TEST_CLASS = Base32BytesPadded
    ENCODED_VALUE = 'KRUGK43FEBQXEZJAMJ4XIZLTEE======'
    SCHEMA_FORMAT = 'base32'
    VALUES: ClassVar[Mapping[str, Sequence[Any]]] = {
        'fail': [
            'KRUGK43FEBQXEZJAMJ4XIZLTE9',  # Invalid character
            'KRUGK43FEBQXEZJAMJ4XIZLTEE',  # Bad padding
        ],
        'good': [
            b'These are bytes!',
            Base32Bytes(b'These are bytes!'),
            'KRUGK43FEBQXEZJAMJ4XIZLTEE======',
            'krugk43febqxezjamj4xizltee======',
        ],
    }


class Base64PaddedEncoder(Base64Encoder):
    trim_padding: ClassVar[bool] = False


Base64BytesPadded = Annotated[bytes, EncodedBytes(encoder=Base64PaddedEncoder)]


@inject_test_values
class TestBase64(BaseEncoderTest):
    TEST_CLASS = Base64Bytes
    ENCODED_VALUE = 'VGhlc2UgYXJlIGJ5dGVzIQ'
    SCHEMA_FORMAT = 'base64'
    VALUES: ClassVar[Mapping[str, Sequence[Any]]] = {
        'good': [
            b'These are bytes!',
            Base64Bytes(b'These are bytes!'),
            'VGhlc2UgYXJlIGJ5dGVzIQ==',
            'VGhlc2UgYXJlIGJ5dGVzIQ',
        ],
        'fail': [
            '=Ghlc2UgYXJlIGJ5dGVzIQ',
        ],
    }


@inject_test_values
class TestBase64Padded(BaseEncoderTest):
    TEST_CLASS = Base64BytesPadded
    ENCODED_VALUE = 'VGhlc2UgYXJlIGJ5dGVzIQ=='
    SCHEMA_FORMAT = 'base64'
    VALUES: ClassVar[Mapping[str, Sequence[Any]]] = {
        'good': [
            b'These are bytes!',
            Base64Bytes(b'These are bytes!'),
            'VGhlc2UgYXJlIGJ5dGVzIQ==',
        ],
        'fail': [
            'VGhlc2UgYXJlIGJ5dGVzIQ',
            '=Ghlc2UgYXJlIGJ5dGVzIQ',
        ],
    }



from __future__ import annotations

import asyncio
import logging
import secrets
from functools import partial
from ipaddress import IPv4Address

import pytest
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from pydantic import ValidationError

from aiostem import Controller
from aiostem.command import CommandHsFetch
from aiostem.event import (
    EventNetworkLiveness,
    EventSignal,
    EventStatusClient,
    EventUnknown,
    event_from_message,
)
from aiostem.exceptions import (
    CommandError,
    ControllerError,
    CryptographyError,
    ReplyStatusError,
)
from aiostem.structures import (
    CircuitPurpose,
    LongServerName,
    OnionServiceKeyStruct,
    StreamCloseReasonInt,
)
from aiostem.utils import Message

# All test coroutines will be treated as marked for asyncio.
pytestmark = pytest.mark.asyncio


@pytest.fixture
def tmp_cookie_data():
    return secrets.token_bytes(32)


@pytest.fixture
def tmp_cookie_path(tmp_path, tmp_cookie_data):
    cookie_path = tmp_path / 'cookiefile'
    with open(cookie_path, 'wb') as fp:
        fp.write(tmp_cookie_data)
    return cookie_path


class TestController:
    """Various tests around the main controller."""

    async def test_initialized(self, controller_raw):
        assert controller_raw.authenticated is False
        assert controller_raw.connected is False
        assert controller_raw.entered is False

    async def test_exit_not_entered(self, controller_raw):
        assert controller_raw.entered is False
        await controller_raw.__aexit__(None, None, None)
        assert controller_raw.entered is False

    async def test_entered(self, controller_unauth):
        assert controller_unauth.authenticated is False
        assert controller_unauth.connected is True
        assert controller_unauth.entered is True

    async def test_already_entered(self, controller_unauth):
        with pytest.raises(RuntimeError, match='Controller is already entered'):
            await controller_unauth.__aenter__()

    async def test_controller_reuse(self, controller_raw):
        for _ in range(3):
            assert controller_raw.entered is False
            async with controller_raw:
                assert controller_raw.entered is True

    async def test_not_entered_from_path(self):
        controller = Controller.from_path('/run/tor/not_a_valid_socket.sock')
        with pytest.raises(FileNotFoundError, match='No such file'):
            await controller.__aenter__()
        assert controller.connected is False

    async def test_not_entered_from_port(self):
        controller = Controller.from_port('qweqwe', 9051)
        assert controller.connected is False

        with pytest.raises(ControllerError, match='Controller is not connected'):
            await controller.protocol_info()

    async def test_unauth_protoinfo(self, controller_unauth):
        res1 = await controller_unauth.protocol_info()
        res2 = await controller_unauth.protocol_info()
        assert res1 == res2

    async def test_authenticate_with_null(self, controller_unauth):
        controller_unauth.enabled_auth_methods = {'NULL'}
        reply = await controller_unauth.authenticate()
        assert reply.is_error is True

    async def test_authenticate_with_safecookie(
        self,
        controller_unauth,
        tmp_cookie_data,
        tmp_cookie_path,
    ) -> None:
        controller_unauth.enabled_auth_methods = {'SAFECOOKIE'}
        controller_unauth.auth_cookie_data = tmp_cookie_data
        controller_unauth.auth_cookie_file = tmp_cookie_path
        reply = await controller_unauth.authenticate()
        assert reply.is_error is True

    async def test_authenticate_with_cookie(
        self,
        controller_unauth,
        tmp_cookie_data,
        tmp_cookie_path,
    ):
        controller_unauth.enabled_auth_methods = {'COOKIE'}
        controller_unauth.auth_cookie_data = tmp_cookie_data
        controller_unauth.auth_cookie_file = tmp_cookie_path
        reply = await controller_unauth.authenticate()
        assert reply.is_error is True

    async def test_authenticate_with_unknown_auth(self, controller_unauth):
        controller_unauth.enabled_auth_methods = {'UNKNOWN'}
        with pytest.raises(ControllerError, match='No compatible authentication method found'):
            await controller_unauth.authenticate()

    async def test_authenticate_no_password(self, controller_unauth):
        with pytest.raises(FileNotFoundError, match='No such file'):
            await controller_unauth.authenticate()

    @pytest.mark.timeout(2)
    async def test_cmd_auth_challenge(self, controller_unauth):
        res = await controller_unauth.auth_challenge(b'NOT A TOKEN')
        res.raise_for_status()

        with pytest.raises(CryptographyError, match='Server hash provided by Tor is invalid'):
            res.data.raise_for_server_hash_error(b'THIS IS A COOKIE')

        token = res.data.build_client_hash(b'THIS IS A COOKIE')
        assert isinstance(token, bytes), token
        # This is the expected length of the client hash.
        assert len(token) == 32

    @pytest.mark.timeout(2)
    async def test_cmd_auth_challenge_error(self, controller_unauth):
        res = await controller_unauth.auth_challenge(b'')
        with pytest.raises(ReplyStatusError, match='Wrong number of arguments'):
            res.raise_for_status()

    @pytest.mark.timeout(2)
    async def test_cmd_auth_challenge_no_nonce(self, controller_unauth):
        # This means that the nonce is generated locally by `auth_challenge`.
        res = await controller_unauth.auth_challenge()
        res.raise_for_status()
        assert len(res.data.client_nonce) == 32
        assert len(res.data.server_nonce) == 32

    async def test_authenticated_controller(self, controller):
        assert controller.connected
        assert controller.authenticated

    async def test_cmd_attach_stream(self, controller):
        reply = await controller.attach_stream(0, 0)
        assert reply.is_error is True
        assert reply.status == 552
        assert 'Unknown stream' in reply.status_text

    async def test_cmd_close_circuit(self, controller):
        reply = await controller.close_circuit(0)
        assert reply.is_error is True
        assert reply.status == 552
        assert 'Unknown circuit' in reply.status_text

    async def test_cmd_extend_circuit(self, controller):
        reply = await controller.extend_circuit(
            0,
            [
                '$9695DFC35FFEB861329B9F1AB04C46397020CE31~Test1',
                LongServerName(
                    fingerprint=b'\xc3\x07\xda\xa2\r(a\x19\tGS\r~XiD\x9d\x0bis',
                ),
            ],
        )
        assert reply.is_error is True
        assert reply.status == 552
        assert 'No such router' in reply.status_text

    async def test_cmd_post_descriptor(self, controller):
        reply = await controller.post_descriptor('XXX')
        assert reply.is_error is True
        assert reply.status == 554
        assert "Couldn't parse router descriptor" in reply.status_text

    async def test_cmd_set_circuit_purpose(self, controller):
        reply = await controller.set_circuit_purpose(0, CircuitPurpose.GENERAL)
        assert reply.is_error is True
        assert reply.status == 552
        assert 'Unknown circuit' in reply.status_text

    async def test_cmd_close_stream(self, controller):
        reply = await controller.close_stream(0, StreamCloseReasonInt.MISC)
        assert reply.is_error is True
        assert reply.status == 552
        assert 'Unknown stream' in reply.status_text

    async def test_cmd_redirect_stream(self, controller):
        reply = await controller.redirect_stream(0, IPv4Address('127.0.0.1'))
        assert reply.is_error is True
        assert reply.status == 552
        assert 'Unknown stream' in reply.status_text

    async def test_cmd_get_info(self, controller):
        info = await controller.get_info('version')
        assert 'version' in info

    async def test_cmd_get_info_error(self, controller):
        res = await controller.get_info('THIS_IS_AN_INVALID_VALUE')
        with pytest.raises(ReplyStatusError, match='Unrecognized key') as exc:
            res.raise_for_status()
        assert exc.value.code >= 400

    async def test_cmd_get_conf(self, controller):
        info = await controller.get_conf('DormantClientTimeout')
        assert info.get('DormantClientTimeout') == '86400'

    async def test_cmd_load_conf(self, controller):
        info = await controller.get_info('config-text')
        info.raise_for_status()

        resp = await controller.load_conf(info['config-text'])
        resp.raise_for_status()

    async def test_map_address(self, controller):
        reply = await controller.map_address(
            {
                'one.one.one.one': '1.1.1.1',
                'dns.google': '8.8.8.8',
            }
        )
        reply.raise_for_status()
        for item in reply.items:
            item.raise_for_status()
        assert len(reply.items) == 2

        info = await controller.get_info('address-mappings/control')
        info.raise_for_status()

        entries = info['address-mappings/control'].splitlines()
        assert isinstance(entries, list)
        assert len(entries) == 2

    async def test_cmd_onion_client_auth_add(self, controller):
        address = 'aiostem26gcjyybsi3tyek6txlivvlc5tczytz52h4srsttknvd5s3qd'
        reply = await controller.onion_client_auth_add(
            address=address,
            key=X25519PrivateKey.generate(),
        )
        assert reply.is_success is True

    async def test_cmd_onion_client_auth_remove(self, controller):
        address = 'aiostpy74pbvneqehctjqan6242vcxeppxmjkschdpqe7tgn7wve65qd'
        added = await controller.onion_client_auth_add(
            address=address,
            key=X25519PrivateKey.generate(),
        )
        assert added.is_success is True

        removed = await controller.onion_client_auth_remove(address)
        assert removed.is_success is True

    async def test_cmd_onion_client_auth_view(self, controller):
        address = 'aiostpy74pbvneqehctjqan6242vcxeppxmjkschdpqe7tgn7wve65qd'
        added = await controller.onion_client_auth_view(address)
        assert added.is_success is True

    async def test_cmd_reset_conf(self, controller):
        conf = {'MaxClientCircuitsPending': '64'}
        result = await controller.reset_conf(conf)
        assert result.status == 250

        info = await controller.get_conf('MaxClientCircuitsPending')
        assert dict(info.items()) == conf

    async def test_cmd_resolve(self, controller):
        result = await controller.resolve(['one.one.one.one'])
        assert result.status == 250

    async def test_cmd_save_conf(self, controller):
        result = await controller.save_conf()
        with pytest.raises(ReplyStatusError, match='Unable to write configuration to disk'):
            assert result.raise_for_status()

    async def test_cmd_set_conf(self, controller):
        conf = {'MaxClientCircuitsPending': '64'}
        result = await controller.set_conf(conf)
        assert result.status == 250

        info = await controller.get_conf('MaxClientCircuitsPending')
        assert dict(info.items()) == conf

    async def test_cmd_protocol_info(self, controller):
        res1 = await controller.protocol_info()
        res1.raise_for_status()

        assert res1.data.auth_cookie_file is not None
        assert res1.data.protocol_version == 1
        assert isinstance(res1.data.tor_version, str)

    async def test_cmd_protocol_info_read_cookie_file_error(self, controller):
        res1 = await controller.protocol_info()
        with pytest.raises(FileNotFoundError, match='No such file or directory'):
            await res1.read_cookie_file()

    async def test_cmd_hs_fetch_v2_error(self, controller):
        reply = await controller.hs_fetch('tor66sezptuu2nta')
        with pytest.raises(ReplyStatusError, match='Invalid argument'):
            reply.raise_for_status()

    async def test_cmd_hs_fetch_with_servers(self, controller):
        # Enable tracing for all sent commands.
        controller.traces.add('command')

        # This is an invalid argument here, no side effect.
        await controller.hs_fetch(
            'tor66sezptuu2nta',
            servers=(
                '$9695DFC35FFEB861329B9F1AB04C46397020CE31~Test1',
                LongServerName(
                    fingerprint=b'\xc3\x07\xda\xa2\r(a\x19\tGS\r~XiD\x9d\x0bis',
                    nickname=None,
                ),
            ),
        )

        # Check that our trace has our command with our servers.
        assert len(controller.trace_commands) == 1
        command = controller.trace_commands[0]
        assert isinstance(command, CommandHsFetch)
        assert len(command.servers) == 2

    async def test_cmd_hs_post(self, controller):
        address = 'oftestt7ffa4tt7et5wab7xhnzeooavy2xdmn6dtfa4pot7dk4xhviid'
        reply = await controller.hs_post('XXX', address=address)
        assert reply.is_error
        assert reply.status == 554
        assert 'Invalid descriptor' in reply.status_text

    async def test_cmd_drop_guard(self, controller):
        res = await controller.drop_guards()
        assert res.status_text == 'OK'

    async def test_cmd_drop_ownership(self, controller):
        res = await controller.drop_ownership()
        assert res.status_text == 'OK'

    async def test_cmd_drop_timeouts(self, controller):
        res = await controller.drop_timeouts()
        assert res.status_text == 'OK'

    async def test_cmd_take_ownership(self, controller):
        res = await controller.take_ownership()
        assert res.status_text == 'OK'

    @pytest.mark.timeout(2)
    async def test_cmd_quit(self, controller):
        """Also checks that both sync and async callbacks are triggered."""
        evt_sync = asyncio.Event()
        evt_async = asyncio.Event()

        def callback_sync(event, _):
            event.set()

        async def callback_async(event, _):
            event.set()

        cb_sync = partial(callback_sync, evt_sync)
        cb_async = partial(callback_async, evt_async)

        await controller.add_event_handler('DISCONNECT', cb_sync)
        await controller.add_event_handler('DISCONNECT', cb_async)

        reply = await controller.quit()
        reply.raise_for_status()

        await asyncio.wait(
            (
                asyncio.create_task(evt_sync.wait()),
                asyncio.create_task(evt_async.wait()),
            ),
            return_when=asyncio.ALL_COMPLETED,
        )
        assert evt_sync.is_set()
        assert evt_async.is_set()

        await controller.del_event_handler('DISCONNECT', cb_async)
        await controller.del_event_handler('DISCONNECT', cb_sync)

    @pytest.mark.timeout(2)
    async def test_add_onion_existing_key(self, controller):
        reply = await controller.add_onion(
            key=Ed25519PrivateKey.generate(),
            ports=['80,127.0.0.1:80'],
        )
        reply.raise_for_status()
        assert reply.data.key is None

    @pytest.mark.timeout(2)
    async def test_add_onion_generate_best_locally(self, controller):
        reply = await controller.add_onion(key='NEW:BEST', ports=['80,127.0.0.1:80'])
        reply.raise_for_status()
        assert isinstance(reply.data.key, Ed25519PrivateKey)

    @pytest.mark.timeout(2)
    async def test_add_onion_generate_rsa1024_locally(self, controller):
        """Note: this is not supported anymore for a few years now."""
        reply = await controller.add_onion(key='NEW:RSA1024', ports=['80,127.0.0.1:80'])
        with pytest.raises(ReplyStatusError, match='Invalid key type'):
            reply.raise_for_status()

    @pytest.mark.timeout(2)
    async def test_add_onion_generate_on_tor(self, controller):
        reply = await controller.add_onion(
            key='NEW:BEST',
            ports=['80,127.0.0.1:80'],
            client_auth_v3=[
                X25519PublicKey.from_public_bytes(secrets.token_bytes(32)),
                X25519PublicKey.from_public_bytes(secrets.token_bytes(32)),
            ],
            generate_locally=False,
        )
        reply.raise_for_status()
        assert isinstance(reply.data.key, OnionServiceKeyStruct)

    @pytest.mark.timeout(2)
    async def test_add_and_del_onion(self, controller):
        added = await controller.add_onion(key='NEW:BEST', ports=['80,127.0.0.1:80'])
        added.raise_for_status()

        deleted = await controller.del_onion(added.data.address)
        deleted.raise_for_status()
        assert deleted.status_text == 'OK'


class TestControllerEvents:
    """Event related tests for the controller."""

    async def test_cmd_add_event_handler_error(self, controller):
        controller.error_on_set_events = True
        with pytest.raises(ReplyStatusError, match='Triggered by PyTest'):
            await controller.add_event_handler('STATUS_CLIENT', lambda: None)
        assert len(controller.event_handlers) == 0

    async def test_cmd_del_event_handler_error(self, controller):
        def callback(_):
            pass

        await controller.add_event_handler('STATUS_CLIENT', callback)
        assert len(controller.event_handlers) == 1

        controller.error_on_set_events = True

        with pytest.raises(ReplyStatusError, match='Triggered by PyTest'):
            await controller.del_event_handler('STATUS_CLIENT', callback)
        assert len(controller.event_handlers) == 1

    async def test_cmd_subscribe_bad_event(self, controller):
        with pytest.raises(CommandError, match="Unknown event 'INVALID_EVENT'"):
            await controller.add_event_handler('INVALID_EVENT', lambda: None)

    async def test_signal_bad_signal(self, controller):
        with pytest.raises(ValidationError, match='Input should be '):
            await controller.signal('INVALID_SIGNAL')

    async def test_bad_event_received(self, controller, caplog):
        message = Message(status=250, header='HELLO WORLD')
        with caplog.at_level(logging.ERROR, logger='aiostem'):
            await controller.push_event_message(message)
        assert 'Unable to handle a received event.' in caplog.text

    @pytest.mark.timeout(2)
    async def test_event_handler_with_error(self, controller, caplog):
        """Check that we log and survive with a broken handler."""

        def cb_sync(event, _):
            event.set()
            msg = 'Unexpected type XXX in YYY.'
            raise TypeError(msg)

        evt_sync = asyncio.Event()
        callback = partial(cb_sync, evt_sync)
        await controller.add_event_handler('DISCONNECT', callback)

        with caplog.at_level(logging.ERROR, logger='aiostem'):
            reply = await controller.quit()
            reply.raise_for_status()
            await evt_sync.wait()

        assert evt_sync.is_set()
        assert "Error while handling callback for 'DISCONNECT'" in caplog.text

    async def test_del_non_existant_event_handler(self, controller):
        assert len(controller.event_handlers) == 0
        await controller.del_event_handler('DISCONNECT', lambda: None)
        assert len(controller.event_handlers) == 0

    @pytest.mark.timeout(2)
    async def test_event_network(self, controller):
        loop = asyncio.get_running_loop()
        future = loop.create_future()

        def on_network_event(fut, event):
            fut.set_result(event)

        callback = partial(on_network_event, future)
        await controller.add_event_handler('NETWORK_LIVENESS', callback)

        message = Message(status=650, header='NETWORK_LIVENESS UP')
        await controller.push_event_message(message)

        evt = await asyncio.ensure_future(future)
        assert isinstance(evt, EventNetworkLiveness)
        assert evt.status == 'UP'

    async def test_unknown_event(self):
        evt = event_from_message(Message(status=650, header='SPECIAL_EVENT'))
        assert isinstance(evt, EventUnknown)

    @pytest.mark.timeout(2)
    async def test_event_status_client(self, controller):
        loop = asyncio.get_running_loop()
        future = loop.create_future()

        def on_status_event(fut, event):
            fut.set_result(event)

        callback = partial(on_status_event, future)
        await controller.add_event_handler('STATUS_CLIENT', callback)

        message = Message(
            header='STATUS_CLIENT NOTICE BOOTSTRAP TAG=done SUMMARY="Done" PROGRESS=100',
            status=650,
        )
        await controller.push_event_message(message)

        evt = await asyncio.ensure_future(future)
        assert isinstance(evt, EventStatusClient)
        assert evt.action == 'BOOTSTRAP'
        assert evt.severity == 'NOTICE'
        assert evt.arguments.progress == 100

    @pytest.mark.timeout(2)
    async def test_event_signal(self, controller):
        loop = asyncio.get_running_loop()
        future = loop.create_future()

        def on_signal_event(fut, event):
            fut.set_result(event)

        callback = partial(on_signal_event, future)
        await controller.add_event_handler('SIGNAL', callback)

        res = await controller.signal('RELOAD')
        assert res.status_text == 'OK'

        evt = await asyncio.ensure_future(future)
        assert isinstance(evt, EventSignal)
        assert evt.signal == 'RELOAD'
